""" 
 * ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *      the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
"""

import mod_snake
import string
import sys

_myname = "SnakeAuth"

class SimpleAuth:
    "A 'simple' authentication plugin module"

    def __init__(self):
        pass
        
    def get_auth_prefixes(self):
        return [ 'SimpleFile',
                 'SimpleGroup',
               ]

    def validate_user_pw(self, svr, user, password, auth_prefix, authdata):
        if auth_prefix == 'SimpleFile':
            try:
                lines = open(authdata, 'r').readlines()
            except:
                mod_snake.ap_log_error(mod_snake.APLOG_ERR, svr,
                                       "Could not open password file: %s" % (
                                             authdata ))
                return mod_snake.HTTP_INTERNAL_SERVER_ERROR

            lineno = 0
            for line in lines:
                lineno = lineno + 1
                if len(line) == 0 or line[0] == '#':
                    continue
                
                x = string.split(line, ':')
                if len(x) < 2:
                    mod_snake.ap_log_error(mod_snake.APLOG_WARNING, svr,
                                           "Error on line %d of password file: %s" % (lineno, authdata))
                    continue
                    
                if len(x) == 2 and x[0] == user and x[1] == password:
                    return mod_snake.OK

        return mod_snake.DECLINED

    def validate_authorization(self, svr,user, auth_prefix, authdata, require):
        if require[0] == 'valid-user':
            return mod_snake.OK

        if require[0] == 'user' and user in require:
            return mod_snake.OK
        
        if require[0] == 'group' and auth_prefix == 'SimpleGroup':
            try:
                lines = open(authdata, 'r').readlines()
            except:
                mod_snake.ap_log_error(mod_snake.APLOG_ERR, svr,
                                       "Could not open group file: %s" % (
                                             authdata ))
                return mod_snake.HTTP_INTERNAL_SERVER_ERROR

            lineno = 0
            for line in lines:
                lineno = lineno + 1
                if len(line) == 0 or line[0] == '#':
                    continue

                x = string.split(line, ':')
                if len(x) != 2:
                    mod_snake.ap_log_error(mod_snake.APLOG_WARNING, svr,
                             "Error on line %d of group file: %s" % (lineno,
                                                                     authdata))
                    continue

                members = string.split(x[1])
                if x[0] in require[1:] and user in members:
                    return mod_snake.OK
        
        return mod_snake.DECLINED
    
class SnakeAuth:
    "A mod_snake pluggable authentication module"

    def create_dir_config(self, path):
        return { "path"     : path,
                 "authinfo" : [],
               }
    def auth_checker(self, per_dir, r):
        methnum = r.method_number
        method_restricted = 0
        for require in r.requires():
            meth_valid = require.method_mask & (1 << methnum)
            if not meth_valid:
                continue
            
            method_restricted = 1

            req_args = mod_snake.ap_getwords_conf(require.requirement)
            for auth_prefix, authdata in per_dir['authinfo']:
                auth_obj = self.auth_modules[auth_prefix]
                rres = auth_obj.validate_authorization(self.server, r.user,
                                                       auth_prefix, authdata,
                                                       req_args)
                if rres != mod_snake.DECLINED:
                    return rres

        if not method_restricted:
            return mod_snake.OK

        r.note_basic_auth_failure()
        return mod_snake.HTTP_UNAUTHORIZED
        
    def check_user_id(self, per_dir, r):
        if not per_dir.has_key('authinfo'):
            return mod_snake.DECLINED
        
        for auth_prefix, authdata in per_dir['authinfo']:
            auth_obj = self.auth_modules[auth_prefix]
            pwinfo = r.get_basic_auth_pw()
            if pwinfo[0] != mod_snake.OK:
                return pwinfo[0]
            
            res = auth_obj.validate_user_pw(self.server, r.user, pwinfo[1],
                                            auth_prefix, authdata)
            if res == mod_snake.OK:
                return mod_snake.OK
            elif res != mod_snake.DECLINED:
                return res
                
        r.note_basic_auth_failure()
        return mod_snake.HTTP_UNAUTHORIZED

    def cmd_SnakeAuth(self, per_dir, authtype):
        "an authtype:authdata pair"

        x = string.split(authtype, ":")
        if len(x) != 2:
            return "SnakeAuth takes an authtype:authdata pair"

        if not self.auth_modules.has_key(x[0]):
            return "AuthType '%s' is not associated with any authmodule" % x[0]
        
        per_dir["authinfo"].append((x[0], x[1]))
        
    def cmd_SnakeAuthTypeAdd(self, per_dir, newmod):
        "an authmodule.authclass pair"

        # Attempt to import the module and instantiate the authclass
        x = string.split(newmod, '.')
        if len(x) != 2:
            return "Invalid argument .. must be authmodule.authclass"

        mod = __import__(x[0], None, None, None)
        obj = mod.__dict__[x[1]]()

        prefixes = obj.get_auth_prefixes()
        if type(prefixes) != type([]):
            return "AuthModule %s returned bad prefixes: %s" % (newmod,
                                                                type(prefixes))
        for prefix in obj.get_auth_prefixes():
            if self.auth_modules.has_key(prefix):
                return "the auth-prefix '%s' has already been defined" % prefix
            self.auth_modules[prefix] = obj
                             
    def __init__(self, module):
        sys.stderr.write('Loading module %s\n' % _myname)
	directives = { "SnakeAuth"        : (mod_snake.OR_ALL, mod_snake.TAKE1,
                                             self.cmd_SnakeAuth),
                       "SnakeAuthTypeAdd" : (mod_snake.RSRC_CONF,
                                             mod_snake.TAKE1,
                                             self.cmd_SnakeAuthTypeAdd),
	}
	module.add_directives(directives)

	hooks = { "create_dir_config"    : self.create_dir_config,
		  "check_user_id"        : self.check_user_id,
		  "auth_checker"         : self.auth_checker,
		}
	      
        for hook in hooks.keys():
	    module.add_hook(hook, hooks[hook])

	self.server  = module.server
        self.auth_modules = {}
        
print "%s loaded" % _myname
