"""
 * ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *      the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
"""

import mod_snake
import os
import rfc822
import string
import sys
import StringIO
import time
import traceback

_myname = "SnakeCGI"

class SnakeCGIPseudoFile:
    "A file which handles communication between our Python CGIs and \
    the connection request object.  This allows us to buffer up data \
    which we think are headers, set them in the headers_out, and use \
    the correct Apache API calls to send the response"

    def __init__(self, request):
        self.connection  = request.connection
        self.request     = request
        self.inheaders   = 1
        self.headerstr   = ""
        
    def read(self, *bytes):
        return apply(self.connection.read, bytes)

    def write(self, str):
        # The connection write function returns a tuple: (success, nwrote)
        # Since these are CGIs, we want to ensure that we write everything
        # Similar to how regular Python does it with files.
        # The catch here is that at the beginning, we want to filter things
        # which appear to be header information.  When it looks like headers
        # have stopped, we send data regular stylee.

        if self.inheaders:
            self.headerstr = self.headerstr + str
            
            # Are we done with header-fetching?
            idx = string.find(self.headerstr, '\n\n')
            if(idx == -1):
                return

            dummyfile = StringIO.StringIO(self.headerstr[0:idx])
            str = self.headerstr[idx:]  
            headers = rfc822.Message(dummyfile)
            headers_out = self.request.headers_out
            # Possibly merge headers? XXX
            for key in headers.keys():
                headers_out[key] = headers[key]

                # Special case for a few headers, as we need to set data
                # in the request_rec based on these

                if string.lower(key) == 'content-type':
                    self.request.content_type = headers[key]
                elif string.lower(key) == 'status':
                    # Take the number, not the string rep
                    self.request.status = int(string.split(headers[key])[0])
                    # XXX Also should allow status-line modifications? Yech
                
            self.request.send_http_header()
            self.inheaders = 0

        nwrote = 0
        succ = mod_snake.APR_SUCCESS
        while nwrote != len(str) and succ == mod_snake.APR_SUCCESS:
            succ, just_wrote = self.connection.write(str)
            nwrote = nwrote + just_wrote
        
class SnakeCGIObject:
    "An object which deals with the actual checking && loading of data \
    from disk, and the namespace maintanence"
     
    def load(self):
        "load():  Load the CGIObject on disk into this object.  The data \
        is compiled and the file-mtime is saved for future usage"
	fil = open(self.filename, 'r')
	x = fil.read()
	self.compiledata = compile(x, self.filename, 'exec')
	self.statinfo   = os.stat(self.filename)
	self.namespace  = {}

    def is_outdated(self):
        "is_outdated():  Returns true of the object has a version more \
        recent on disk, false if not."
	statinfo = os.stat(self.filename)
	if statinfo[-2] != self.statinfo[-2]:
	    return 1
	return 0

    def __init__(self, filename):
	self.filename = filename
	self.load()

    def run(self, request_rec):
        "run(request_rec):  Runs the CGI, sending the output to the \
        connection request.  Returns exception info on failure, \
        None on success"

	# Set the stdout of the CGI, and save ours
        pseudofile = SnakeCGIPseudoFile(request_rec)
	saveout, savein = sys.stdout, sys.stdin
	sys.stdout, sys.stdin = pseudofile, pseudofile
	try:
	    eval(self.compiledata, self.namespace, self.namespace)
	except:
	    sys.stdout, sys.stdin = saveout, savein
	    etype, value, tb = sys.exc_info()
	    return traceback.format_exception(etype, value, tb, None)
	
        sys.stdout, sys.stdin = saveout, savein
        if pseudofile.inheaders:
            return "Malformed header section"
        
	return None

class SnakeCGI:
    def cmd_SnakeCGILogFile(self, per_dir, path):
	"Specify a LogFile for output errors from CGI scripts"
        self.logfile = mod_snake.ap_server_root_relative(path)
    
    def create_dir_config(self, path):
        return None
        
    def content_handler(self, per_dir, r):
	if r.handler != 'snakecgi':
	    return mod_snake.DECLINED

	filename = r.filename
	cgis = self.cgis

	if not cgis.has_key(filename) or cgis[filename].is_outdated():
	    try:
		self.cgis[filename] = SnakeCGIObject(filename)
	    except IOError:
		return mod_snake.HTTP_NOT_FOUND
	    except:
		etype, value, tb = sys.exc_info()
		tback = traceback.format_exception(etype, value, tb, None)
                tot = string.join(tback[4:], '')
		self.write_log(mod_snake.APLOG_CRIT, 
			       "Error loading script %s\n%s" % (filename, tot))
		return mod_snake.HTTP_INTERNAL_SERVER_ERROR

        r.add_common_vars()
        r.add_cgi_vars()
        os.environ.update(r.subprocess_env)
        os.environ['GATEWAY_INTERFACE'] = "FOOBAR"

	exc = cgis[filename].run(r)
	if exc != None:
            if type(exc) == type(""):
                self.write_log(mod_snake.APLOG_CRIT,
                               "Error running script %s: %s" % (filename, exc))
            else:
                tot = string.join(exc[2:],'')
                self.write_log(mod_snake.APLOG_CRIT, 
                               "Error running script %s\n%s" % (filename, tot))

	return mod_snake.OK

    def write_log(self, level, logstr):
	if self.logfd:
	    self.logfd.write("[%s] - %s\n" % (time.ctime(time.time()), logstr))
	    self.logfd.flush()
	else:
	    mod_snake.ap_log_error(level, self.server, logstr)
	
    def open_logs(self, moduledata):
        if self.logfile:
            self.logfd = open(self.logfile, "a")
	else:
	    self.logfd = None
    
    def __init__(self, module):
	# Configuration directives that we support
	directives = { "SnakeCGILogFile" :(mod_snake.RSRC_CONF,mod_snake.TAKE1,
                                           self.cmd_SnakeCGILogFile),
		     }
	module.add_directives(directives)

	hooks = { "create_dir_config"    : self.create_dir_config,
                  "open_logs"            : self.open_logs,
		  "content_handler"      : self.content_handler,
		}
	      
        for hook in hooks.keys():
	    module.add_hook(hook, hooks[hook])

	self.cgis    = {}
        self.logfile = None
	self.server  = module.server

        for key in os.environ.keys():
            del os.environ[key]
        
    def __del__(self):
	# Perform any cleanup here
	pass

#print "%s loaded" % (_myname,)
