""" 
 * ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *      the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
"""

#
# TODO:  Perhaps speed this up with some C code?
#

import string
import mod_snake
import sys
import os
import traceback
import time

_myname = "SnakeEMB"

class SnakeEMBMinusObject:
    def __init__(self):
	self.data = None

    def gobble(self, filename, str, startidx=0):
	"Gobbles until it reaches a -].  Returns the idx of the ]"
	
	i = startidx
	while i < len(str):
	    if str[i] == '-' and i != len(str) and str[i+1] == ']':
		self.data = str[startidx:i]
		
		self.data = compile(self.data, "<internal_%s>" % filename,
				    "exec")
		return i + 1
	    else:
		i = i + 1

	return -1

    def process(self, namespace, fd):
	exec(self.data, namespace, namespace)


class SnakeEMBPlusObject:
    def __init__(self):
	self.data = None

    def gobble(self, filename, str, startidx=0):
	"Gobbles until it reaches a +].  Returns the idx of the ]"
	
	i = startidx
	while i < len(str):
	    if str[i] == '+' and i != len(str) and str[i+1] == ']':
		self.data = str[startidx:i]
		
		self.data = compile(self.data, "<internal_%s>" % filename,
				    "eval")
		return i + 1
	    else:
		i = i + 1

	return -1

    def process(self, namespace, fd):
	x = eval(self.data, namespace, namespace)
	if type(x) == type(""):
	    str = x
	else:
	    str = `x`
	fd.write(str)
	
	
class SnakeEMBTextObject:
    def __init__(self):
	self.data = []

    def gobble(self, filename, str, startidx=0):
	"Gobbles chars until one of it's children can eat, or it is done. \
	the IDX of the last char eaten will be returned"
	
	startrange = startidx
	i = startrange
	while i < len(str):
	    if str[i] == '[' and i != (len(str)-1) and (str[i+1] == '+' or
							str[i+1] == '-'):
		# Store our current yum-yums
		self.data.append(str[startrange:i])
		
		# Grab a new object to process
		if str[i+1] == '+':
		    new_obj = SnakeEMBPlusObject()
		else:
		    new_obj = SnakeEMBMinusObject()
		i = startrange = new_obj.gobble(filename, str, i + 2) + 1
		if startrange == -1:
		    return -1

		self.data.append(new_obj)
	    else:
		i = i + 1

	self.data.append(str[startrange:i])

    def process(self, namespace, fd):
	data = self.data;
	for i in range(len(data)):
	    if type(data[i]) == type(""):
		fd.write(data[i])
	    else:
		data[i].process(namespace, fd)


class SnakeEMBProcesser:
    "An object which is used to store parse information for each requested \
    HTML object"
    
    def load(self):
	"Load up information from our on-disk HTML file"
	
	self.filetext  = open(self.filename, 'r').read()
	self.filestat  = os.stat(self.filename)
	self.txtobject = None
	self.namespace = {}

    def is_outdated(self):
	"is_outdated():  Returns true of the object has a version more \
	recent on disk, false if not."

	statinfo = os.stat(self.filename)
	if statinfo[-2] != self.filestat[-2]:
	    return 1
	return 0

    def rebuild_objects(self):
	"Run through the file text, checking for the tags to create new \
	objects"
	    
	self.txtobject = SnakeEMBTextObject()
	if self.txtobject.gobble(self.filename, self.filetext):
	    raise "Uh oh!"

    def process(self, conn_rec):
	if self.txtobject == None or self.is_outdated():
	    self.txtobject = None;
	    self.load()
	    self.rebuild_objects()

	self.txtobject.process(self.namespace, conn_rec)

    def __init__(self, filename):
	self.filename = filename
	self.load()


class SnakeEMB:
    "An embedded Python processer"

    def cmd_SnakeEMBLogFile(self, per_dir, path):
	"cmd_SnakeEMBLogFile: path to the error log file"

	self.logfile = mod_snake.ap_server_root_relative(path)

    def cmd_SnakeEMB(self, per_dir, on_or_off):
	"cmd_SnakeEMB: enable or disable embedded Python"

	per_dir["enabled"] = on_or_off

    def cmd_SnakeEMBErrSend(self, per_dir, on_or_off):
	"cmd_SnakeEMBErrSend: Send traceback errors out the HTML pipe instead \
	of simply sending a failure message"

	per_dir["errsend"] = on_or_off

    def cmd_SnakeEMBCache(self, per_dir, on_or_off):
	"cmd_SnakeEMBCache: enable or disable caching of embedded Python"

	per_dir["cache"] = on_or_off

    def create_dir_config(self, path):
	return { "cache"   : 0,
	         "enabled" : 0, 
		 "errsend" : 0,
		 "path"    : path }
        
    def content_handler(self, per_dir, r):
	if not per_dir["enabled"] or r.content_type != 'text/html':
	    return mod_snake.DECLINED
	
	filename = r.filename
	cache = self.cache
	if not per_dir["cache"] or not cache.has_key(filename):
	    try: 
		emb_obj = SnakeEMBProcesser(filename)
	    except IOError:
		return mod_snake.HTTP_NOT_FOUND
	    except:
		return mod_snake.HTTP_INTERNAL_SERVER_ERROR

            r.send_http_header()

	    try:
		emb_obj.process(r.connection)
	    except:

		etype, value, tb = sys.exc_info()
		tback = traceback.format_exception(etype, value, tb, None)
                tot = string.join(tback[6:],'')

		if not per_dir["errsend"]:
		    r.connection.write("Internal SnakeEMB error\n")
		else: 
		    r.connection.write('<PRE>' + tot + '</PRE>')

		self.write_log("Error in embedded file %s\n:%s" % 
			       (filename, tot))
		    
		return mod_snake.OK

	    if per_dir["cache"]:
		cache[filename] = emb_obj
	    return mod_snake.OK
	
	r.connection.write('HTTP/1.1 200 OK\n')  # XXX
	r.connection.write('Content-Type: text/html\n\n')
	#r.connection.write("<PRE>Cache Hit\n</PRE>")
	emb_obj = cache[filename]
	try:
	    emb_obj.process(r.connection)
	except:
	    etype, value, tb = sys.exc_info()
	    tback = traceback.format_exception(etype, value, tb, None)
            tot = string.join(tback[5:], '')
	    if not per_dir["errsend"]:
		r.connection.write("Internal SnakeEMB error\n")
	    else: 
		r.connection.write('<PRE>' + tot + '</PRE>')

	    self.write_log("Error in embedded file %s\n:%s" % 
	       	       (filename, tot))
	    del cache[filename]

	return mod_snake.OK

    def write_log(self, logstr):
	if self.logfd:
	    self.logfd.write("[%s] - %s\n" % (time.ctime(time.time()), logstr))
	    self.logfd.flush()
	else:
	    mod_snake.ap_log_error(mod_snake.APLOG_WARNING, self.server,
				   logstr)

    def open_logs(self, moduledata):
	if self.logfile:
	    self.logfd = open(self.logfile, "a")
	else:
	    self.logfd = None
    
    def __init__(self, module):
	# Configuration directives that we support
	directives = { "SnakeEMB" :(mod_snake.OR_ALL, mod_snake.FLAG,
				    self.cmd_SnakeEMB),
		       "SnakeEMBCache" :(mod_snake.OR_ALL, mod_snake.FLAG,
					 self.cmd_SnakeEMBCache),
		       "SnakeEMBErrSend" :(mod_snake.OR_ALL, mod_snake.FLAG,
					 self.cmd_SnakeEMBErrSend),
		       "SnakeEMBLogFile" :(mod_snake.RSRC_CONF,mod_snake.TAKE1,
					   self.cmd_SnakeEMBLogFile),
	}
	module.add_directives(directives)

	hooks = { "create_dir_config"    : self.create_dir_config,
                  "open_logs"            : self.open_logs,
		  "content_handler"      : self.content_handler,
		}
	      
        for hook in hooks.keys():
	    module.add_hook(hook, hooks[hook])

	self.server  = module.server
	self.cache   = {}             # Serverwide cache, even across dirs
	                              # so .htaccess EMBCache is meaningful
	self.logfile = None

print "%s loaded" % (_myname,)
