/* ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *     the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include "mod_snake.h"

API_VAR_EXPORT module snake_module;
static int Python_Loaded = 0;   /* Indicates whether or not it has been
                                   initialized yet.  This goes across
                                   all configs, so we need a global */
static PyThreadState *Main_Threadstate = NULL;

static
void mod_snake_initialize_python(){
  if(Python_Loaded) 
    return;

  Python_Loaded = 1;
  Py_Initialize();
#ifdef WITH_THREAD
  PyEval_InitThreads();  /* Acquires a global interp lock for us */
#endif
  Main_Threadstate = PyThreadState_Get();
  PyEval_SaveThread();   /* Release the main thread && Lock      */
}

/*
 * mod_snake_post_cfg_cb:  Post configuration callback.  Setup per-module
 *                         type things such as thread locks.
 */

#ifndef APACHE13
static 
void mod_snake_post_cfg_cb(ap_pool_t *p, ap_pool_t *plog,
                           ap_pool_t *ptemp, server_rec *s)
{
}
#endif

/*
 * unload_dir_cfg:  Unload a per-directory configuration entity.  The per-dir
 *                  config contains callback data for all modules which have
 *                  a registered directive underneath that directory.
 *
 * Arguments:       vdata = ModSnakeDirCfg directory configuration.
 *
 *
 * Side Notes:      XXX -- If the directory config is destroyed AFTER the
 *                         server it was created within, then some data
 *                         will be orphaned.  A better explaination:  The
 *                         module thread + interp data is stored at the server
 *                         config level.  When directory configs are created,
 *                         the callback data is really associated with these
 *                         interps within the servercfg.  Therefore when a
 *                         servercfg destroys the interp, and then the
 *                         dir_cfg wants to destroy data that it kept from
 *                         that interp, it can't because it has been orphaned.
 */
static
#ifdef APACHE13
void 
#else
ap_status_t
#endif
unload_dir_cfg(void *vdata){
  ModSnakeDirCfg    *cfg         = vdata;
  ap_array_header_t *dir_data    = cfg->dir_config_data;
  int i;

  for(i=0; i< dir_data->nelts; i++){
    ModSnakeCallbackData *cdata = (ModSnakeCallbackData *)dir_data->elts + i;

    /* Un-threadsafe? - XXX */
    if(cdata->module->main_state == NULL){
      fprintf(stderr, "Orphaned directory config data in module: %p\n", 
	      cdata->module);
      continue;
    }
    PyEval_RestoreThread(cdata->module->main_state);       /* Lock */
    Py_DECREF(cdata->data);
    PyEval_SaveThread();                                   /* Unlock */
  }
#ifdef APACHE13
  return;
#else
  return APR_SUCCESS;
#endif
}

/*
 * unload_server_cfg:  Unload a server configuration entity.  All modules
 *                     created within this server entity will be unloaded, and
 *                     all other Python objects within this server will be 
 *                     decref'd.
 *
 * Arguments:          vdata = Server configuration entity
 */

static
#ifdef APACHE13
void 
#else
ap_status_t
#endif
unload_server_cfg(void *vdata){
  ModSnakeServerCfg *cfg         = vdata;
  int i;


  for(i=0; i< cfg->command_recs->nelts; i++){
    ModSnakeIntDirective *intdir = 
      (ModSnakeIntDirective *)cfg->command_recs->elts + i;
    
    PyEval_RestoreThread(intdir->pymod->main_state);  /* Lock */
    Py_DECREF(intdir->funcptr);  
    PyEval_SaveThread();                              /* Unlock */
  }

  for(i=0; i< cfg->modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)cfg->modules->elts + i;
    
    PyEval_RestoreThread(pymod->main_state);  /* Lock   */
    mod_snake_unload_module(pymod);
    PyEval_SaveThread();                      /* Unlock */
  }
#ifdef APACHE13
  return;
#else
  return APR_SUCCESS;
#endif
}


#if 0
static void *mod_snake_merge_server_cfg_cb(ap_pool_t *p, void *base,
					    void *newc)
{
  fprintf(stderr, "SERVER MERGE: %p + %p\n", base, newc);
  return newc;
}

#endif

static void *mod_snake_create_server_cfg_cb(ap_pool_t *p, server_rec *s){
  ModSnakeServerCfg *res = ap_pcalloc(p, sizeof(*res));
  res->cfgpool        = p;
  res->moduledirs     = ap_make_array(p, 50, sizeof(char **));
  res->modules        = ap_make_array(p, 50, sizeof(ModSnakePyMod));
  res->command_recs   = ap_make_array(p, 50, sizeof(ModSnakeIntDirective));
  res->server         = s;
  ap_register_cleanup(p, res, unload_server_cfg, ap_null_cleanup);
  return res;
}

static void *mod_snake_create_dir_cfg_cb(ap_pool_t *p, char *dir)
{
  ModSnakeDirCfg *res;

  mod_snake_initialize_python();
  res = ap_pcalloc(p, sizeof(*res));
  res->cfgpool         = p;
  res->dir_config_data = ap_make_array(p, 50, sizeof(ModSnakeCallbackData));
  res->path            = ap_pstrdup(p, dir);
  res->parent          = NULL;
  ap_register_cleanup(p, res, unload_dir_cfg, ap_null_cleanup);
  return res;
}

/*
 * mod_snake_merge_dir_cfg_cb:  Callback for merging of per-dir configuration
 *                              data.  Here, we loop through the modules
 *                              involved in the merge.  Those that want
 *                              to be merged (i.e. they have hooked the
 *                              handler), we call the handler, and store their
 *                              result.  Those who do not want to be merged,
 *                              we simply add them to this new dir-cfg and
 *                              incref their respective callback data.
 */

static void *mod_snake_merge_dir_cfg_cb(ap_pool_t *p, void *base_vdata,
					void *newc_vdata)
{
  ModSnakeDirCfg *base = base_vdata, *newc = newc_vdata, *res;
  ap_array_header_t *dcfg_base;
  int i;

  res         = mod_snake_create_dir_cfg_cb(p, newc->path);
  res->parent = base;

  dcfg_base = base->dir_config_data;
  for(i=0; i < dcfg_base->nelts; i++){
    ModSnakeCallbackData *bdata = (ModSnakeCallbackData *)dcfg_base->elts + i;
    ap_array_header_t *dcfg_new = newc->dir_config_data;
    int j;
    
    for(j=0; j < dcfg_new->nelts; j++){
      ModSnakeCallbackData *ndata = (ModSnakeCallbackData *)dcfg_new->elts + j;
      ModSnakeCallbackData *new_resdata;

      if(ndata->module == bdata->module){
	PyObject *hook, *data;
	ModSnakePyMod *pymod = ndata->module;

	if((hook = pymod->hooks[SNAKE_HOOK_MERGE_DIR_CONFIG]) == NULL){
	  data = ndata->data;
	  Py_INCREF(data);
	} else {
	  data = mod_snake_run_hook_merge_dir_config(pymod, 
						     pymod->parentsvr->server,
						     bdata->data, ndata->data,
						     hook);
	  if(data == NULL){
	    fprintf(stderr, "Error calling merge_dir for: %s\n",
		    pymod->name);
	    continue;
	  }
	}
	new_resdata         = ap_push_array(res->dir_config_data);
	new_resdata->data   = data;
	new_resdata->module = pymod;
      }
    }      
  }
  return res;
}


#ifdef APACHE13
static int file_exists(const char *path, ap_pool_t *p){
  struct stat statbuf;
  int err;

  if((err = stat(path, &statbuf)) == -1)
    return -1;
  
  if(S_ISDIR(statbuf.st_mode))
    return -2;

  return 0;
}
#else
static int file_exists(const char *path, ap_pool_t *p){
  ap_finfo_t finfo;
  ap_status_t err;

  if((err = ap_stat(&finfo, path, p)) != APR_SUCCESS)
    return -1;
  
  if(finfo.filetype == APR_DIR)
    return -2;

  return 0;
}
#endif

/*
 * mod_snake_add_moduledir:  Add a directory to the list of directories
 *                           that contain Python modules for import.
 */

static 
const char *mod_snake_add_moduledir(cmd_parms *cmd, void *dcfg, char *arg) {
  ModSnakeServerCfg *svr_cfg = ap_get_module_config(cmd->server->module_config,
						    &snake_module);
  const char *real_path;
  int stres;

  /* Validate the existance first, so we can give better error msgs */
  real_path = ap_server_root_relative(cmd->pool, arg);
  if((stres = file_exists(real_path, cmd->pool)) == -1)
    return ap_psprintf(cmd->pool, "mod_snake: path '%s' unaccessable", arg);

  if(stres != -2)
    return ap_psprintf(cmd->pool, "mod_snake: path '%s': not a directory", 
                       arg);

  *(char **)ap_push_array(svr_cfg->moduledirs) = ap_pstrdup(cmd->pool, 
							    real_path); 
  return NULL;
}

/*
 * mod_snake_find_module_byname:  Find a module by it's module.classname name.
 *                                
 * Arguments:                     mods = Array of ModSnakePyMod's
 *                                name = module.classname to find.
 *
 * Return values:                 Returns NULL if the module was not found,
 *                                else a pointer to it.
 */

static
ModSnakePyMod *mod_snake_find_module_byname(ap_array_header_t *mods, 
					    const char *name)
{
  int i;

  for(i=0; i < mods->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)mods->elts + i;
    
    if(!strcmp(pymod->name, name)) return pymod;
  }

  return NULL;
}

/*
 * mod_snake_get_dir_cback_data:  Check a directory config to see if a 
 *                                module has had its create_dir_config hook
 *                                called, and subsequent data stored.  If it
 *                                hasn't, call the module's create_dir_config
 *                                hook and store their resultant data.
 *
 * Arguments:                     dir_cfg = Directory config to check
 *                                pymod   = Module to check for callback-fu
 *
 * Side Notes:                    The interp lock must NOT be held when
 *                                calling this function.
 *
 * Return values:                 Returns -1 on an anomaly, else 0
 */

static
int mod_snake_get_dir_cback_data(ModSnakeDirCfg *dir_cfg, ModSnakePyMod *pymod)
{
  ap_array_header_t *dir_data = dir_cfg->dir_config_data;
  ModSnakeCallbackData *cdata;
  PyObject *data, *hook;
  int i;

  if((hook = pymod->hooks[SNAKE_HOOK_CREATE_DIR_CONFIG]) == NULL)
    return 0;
  
  for(i=0; i< dir_data->nelts; i++){
    cdata = (ModSnakeCallbackData *)dir_data->elts + i;
    
    if(cdata->module == pymod)  /* Already have callback data for this fiend */
      return 0;
  }

  data = mod_snake_run_hook_create_dir_config(pymod, pymod->parentsvr->server,
					      dir_cfg, hook);
  if(data == NULL)
    return -1;

  cdata         = ap_push_array(dir_data);
  cdata->data   = data;
  cdata->module = pymod;
  return 0;
}


/*
 * mod_snake_add_module:  Add a module to the list of modules that mod_snake
 *                        will manage.  The module will then be loaded so
 *                        that we can get hooks and determine if the server
 *                        is even going to run.
 *
 * Arguments:             arg1    = Module name to import (as in import foo)
 *                        arg2..n = (optional)  
 *                          failexit = cause a server exit on any kind of
 *                                     failure emitted from a Python module
 *                         
 */

static 
const char *mod_snake_add_module(cmd_parms *cmd, void *dcfg, const char *args)
{
  ModSnakeServerCfg *svr_cfg = ap_get_module_config(cmd->server->module_config,
						    &snake_module);
  ModSnakeDirCfg *dir_cfg    = dcfg;
  ModSnakePyMod *pymod;
  char *module_name, *cp;
  int is_failexit, i;

  mod_snake_initialize_python();  /* Ensure that we are setup */
  if((module_name = ap_getword_conf(cmd->pool, &args)) == NULL)
    return "mod_snake: First argument must be a module name";

  /* Check to see if this module is already handled within our servercfg */
  if(mod_snake_find_module_byname(svr_cfg->modules, module_name) != NULL)
    return "mod_snake: Duplicate module additions within server not allowed";

  /* Check validity of the module.classname */
  if((cp = strchr(module_name, '.')) == NULL)
    return "mod_snake: Module name must be in the form of module.classname";

  is_failexit = 0;
  while(*args){
    char *cp = ap_getword_conf(cmd->pool, &args);
    if(!strcmp(cp, "failexit")) is_failexit = 1;
    else return "mod_snake: Optional parameter must be 'failexit'";
  }

  pymod                            = ap_push_array(svr_cfg->modules);
  pymod->parentsvr                 = svr_cfg;
  pymod->name                      = module_name;
  pymod->module                    = NULL;  
  pymod->mod_obj                   = NULL;
  pymod->interp                    = NULL;
  pymod->main_state                = NULL;
  pymod->valid                     = 0;
  pymod->failexit                  = is_failexit;
  pymod->current_hook              = SNAKE_HOOK_NO_HOOK;
  for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++)
    pymod->hooks[i] = NULL;
  
  /* Called pre-fork, pre-thread, so we shoudln't need to worry about doing
     locking .. but oh well */
  PyEval_RestoreThread(Main_Threadstate); /* Locks */
  if(mod_snake_load_module(pymod, svr_cfg->moduledirs, cmd->server) == -1){
    PyEval_SaveThread();                  /* Unlocks */
    return ap_psprintf(cmd->pool, "mod_snake: Unable to load module %s",
                       module_name);
  }
  PyEval_SaveThread();                    /* Unlocks */

  /* Create the initial per-dir configuration for the module */
  if(mod_snake_get_dir_cback_data(dir_cfg, pymod) == -1) {
    return ap_psprintf(cmd->pool, "mod_snake: Error getting directory info "
		       "from module %s", module_name);
  }
  return NULL;
}

static
void mod_snake_open_logs_cb(ap_pool_t *pconf, ap_pool_t *plog, ap_pool_t *ptmp,
			    server_rec *s)
{
  ModSnakeServerCfg *cfg=ap_get_module_config(s->module_config, &snake_module);
  int i;
  
  for(i=0; i< cfg->modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)cfg->modules->elts + i;

    if(!pymod->hooks[SNAKE_HOOK_OPEN_LOGS]) continue;
    mod_snake_run_hook_open_logs(pymod);
  }
}

#ifdef APACHE13
static void mod_snake_module_init(server_rec *s, pool *p) {
  char *server_string = PACKAGE "/" VERSION;

  mod_snake_open_logs_cb(p, p, p, s);
  ap_add_version_component(server_string);
}
#endif /* APACHE13 */

/* XXX: rmorgan
 * 
 * connection hooks are not available in Apache 1.3.  Remove them.
 */
#ifndef APACHE13
static 
int mod_snake_generic_connection_cb(conn_rec *c, int hook_number){
  ModSnakeServerCfg *cfg=ap_get_module_config(c->base_server->module_config,
					      &snake_module);
  int i, res, hook_type;
  
  hook_type = mod_snake_hooknum_to_type(hook_number);
  for(i=0; i< cfg->modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)cfg->modules->elts + i;
    PyObject *callback = pymod->hooks[hook_number];

    if(!callback) continue;
    res = mod_snake_run_generic_conn_rec(pymod, c, callback, hook_number);
    
    if(hook_type == SNAKE_HOOK_ALL && (res != DECLINED && res != OK))
      return res;

    if(hook_type == SNAKE_HOOK_FIRST && res != DECLINED)
      return res;
  }

  return DECLINED;
}

static 
int mod_snake_pre_connection_cb(conn_rec *c){
  return mod_snake_generic_connection_cb(c, SNAKE_HOOK_PRE_CONNECTION);
}

static 
int mod_snake_process_connection_cb(conn_rec *c){
  return mod_snake_generic_connection_cb(c, SNAKE_HOOK_PROCESS_CONNECTION);
}
#endif

static
int mod_snake_generic_request_cb(request_rec *r, int hook_number)
{
  ModSnakeServerCfg *cfg=ap_get_module_config(r->server->module_config,
					      &snake_module);
  ModSnakeDirCfg *dcfg = ap_get_module_config(r->per_dir_config,&snake_module);
  ModSnakeDirCfg *dirsrch;
  int i, j, res, hook_type;
  
  hook_type = mod_snake_hooknum_to_type(hook_number);
  for(i=0; i< cfg->modules->nelts; i++){
    ModSnakePyMod *pymod = (ModSnakePyMod *)cfg->modules->elts + i;
    PyObject *callback = pymod->hooks[hook_number];

    if(!callback) continue;

    dirsrch = dcfg;
    while(dirsrch != NULL){
      int found = 0;
      ap_array_header_t *dir_cfgs = dirsrch->dir_config_data;

      for(j=0; j < dir_cfgs->nelts; j++){
	ModSnakeCallbackData *cbdata = (ModSnakeCallbackData *)
	                                     dir_cfgs->elts + j;
	if(cbdata->module == pymod){
	  res = mod_snake_run_generic_request_rec(pymod, r, dirsrch, callback, 
						  hook_number);
	  found = 1;
	  break;
	}
      }
      if(found)
	break;
      dirsrch = dirsrch->parent;
    }

    if(hook_type == SNAKE_HOOK_ALL && (res != DECLINED && res != OK))
      return res;

    if(hook_type == SNAKE_HOOK_FIRST && res != DECLINED)
      return res;
  }
  return DECLINED;
}

static 
int mod_snake_header_parser_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_HEADER_PARSER);
}

static 
int mod_snake_post_read_request_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_POST_READ_REQUEST);
}

static 
int mod_snake_log_transaction_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_LOG_TRANSACTION);
}

static 
int mod_snake_translate_name_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_TRANSLATE_NAME);
}

static 
int mod_snake_check_user_id_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_CHECK_USER_ID);
}
static 
int mod_snake_fixups_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_FIXUPS);
}

static 
int mod_snake_type_checker_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_TYPE_CHECKER);
}

static 
int mod_snake_access_checker_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_ACCESS_CHECKER);
}

static 
int mod_snake_auth_checker_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_AUTH_CHECKER);
}

static
int mod_snake_content_handler_cb(request_rec *r){
  return mod_snake_generic_request_cb(r, SNAKE_HOOK_CONTENT_HANDLER);
}

#ifndef APACHE13
static void mod_snake_register_hooks_cb(void)
{
  ap_hook_open_logs(mod_snake_open_logs_cb, NULL, NULL, AP_HOOK_MIDDLE);
  ap_hook_post_config(mod_snake_post_cfg_cb, NULL, NULL, AP_HOOK_MIDDLE);
  ap_hook_pre_connection(mod_snake_pre_connection_cb, NULL, NULL, 
			 AP_HOOK_MIDDLE);
  ap_hook_process_connection(mod_snake_process_connection_cb, NULL, NULL, 
			     AP_HOOK_MIDDLE);
  ap_hook_post_read_request(mod_snake_post_read_request_cb, NULL, 
			    NULL, AP_HOOK_MIDDLE);
  ap_hook_header_parser(mod_snake_header_parser_cb, NULL, NULL,AP_HOOK_MIDDLE);
  ap_hook_log_transaction(mod_snake_log_transaction_cb, NULL, NULL,
			  AP_HOOK_MIDDLE);
  ap_hook_translate_name(mod_snake_translate_name_cb, NULL, NULL,
			 AP_HOOK_MIDDLE);
  ap_hook_check_user_id(mod_snake_check_user_id_cb, NULL, NULL,
			AP_HOOK_MIDDLE);
  ap_hook_fixups(mod_snake_fixups_cb, NULL, NULL,
		 AP_HOOK_MIDDLE);
  ap_hook_type_checker(mod_snake_type_checker_cb, NULL, NULL,
		       AP_HOOK_MIDDLE);
  ap_hook_access_checker(mod_snake_access_checker_cb, NULL, NULL,
			 AP_HOOK_MIDDLE);
  ap_hook_auth_checker(mod_snake_auth_checker_cb, NULL, NULL,
		       AP_HOOK_MIDDLE);
}
#endif

static const command_rec mod_snake_cmds[] = {
{ "SnakeModuleDir", mod_snake_add_moduledir, NULL, RSRC_CONF, ITERATE,
  "Load a module to be used by mod_snake" },
{ "SnakeModule",    mod_snake_add_module,    NULL, RSRC_CONF, RAW_ARGS,
  "Load a module to be used by mod_snake : \
   SnakeModule module.moduleclass [failexit]"},
{ NULL }
};

static handler_rec mod_snake_handlers [] = {
  { "*",            mod_snake_content_handler_cb },
  { NULL },
};

#ifndef APACHE13
API_VAR_EXPORT module snake_module = {
  STANDARD20_MODULE_STUFF,
  mod_snake_create_dir_cfg_cb,    /* create per-directory config structure */
  mod_snake_merge_dir_cfg_cb,     /* merge per-directory config structure  */
  mod_snake_create_server_cfg_cb, /* create per-server config structure    */
  NULL,
  mod_snake_cmds,                 /* command ap_table_t                    */
  mod_snake_handlers,             /* handlers                              */
  mod_snake_register_hooks_cb,    /* register hooks                        */
};
#endif

#ifdef APACHE13
API_VAR_EXPORT module snake_module = {
  STANDARD_MODULE_STUFF,
  mod_snake_module_init,	  /* module initializer */
  mod_snake_create_dir_cfg_cb,    /* per-directory config creator */
  mod_snake_merge_dir_cfg_cb,	  /* dir config merger */
  mod_snake_create_server_cfg_cb, /* server config creator */
  NULL,				  /* server config merger */
  mod_snake_cmds,		  /* command table */
  mod_snake_handlers,             /* [7] list of handlers */
  mod_snake_translate_name_cb,    /* [2] filename-to-URI translation */
  mod_snake_check_user_id_cb,     /* [5] check/validate user_id */
  mod_snake_auth_checker_cb,      /* [6] check user_id is valid *here* */
  mod_snake_access_checker_cb,    /* [4] check access by host address */
  mod_snake_type_checker_cb,      /* [7] MIME type checker/setter */
  mod_snake_fixups_cb,            /* [8] fixups */
  mod_snake_log_transaction_cb,   /* [10] logger */
  mod_snake_header_parser_cb,     /* [3] header parser */
  NULL,               		  /* apache child process initializer */
  NULL,                           /* apache child process exit/cleanup */
  mod_snake_post_read_request_cb  /* [1] post read_request handling */
};						      
#endif

static
ModSnakeIntDirective *
mod_snake_internal_directive_find(ap_array_header_t *cmd_recs,const char *find)
{
  int i;

  for(i=0; i< cmd_recs->nelts; i++){
    ModSnakeIntDirective *int_dir = (ModSnakeIntDirective *)cmd_recs->elts + i;
    
    if(!strcasecmp(int_dir->name, find))
      return int_dir;
  }
  return NULL;
}

static 
const char *mod_snake_setup_internal_directive_call(cmd_parms *cmd,
						    void *per_dir_vdata,
						    PyThreadState **new_state,
						    PyObject **callback_data,
						    PyObject **callback_func,
						    ModSnakePyMod **pymod)
{
  ModSnakeServerCfg *svrcfg;
  ModSnakeDirCfg *dircfg = per_dir_vdata;
  ModSnakeIntDirective *int_dir;
  ap_array_header_t *command_recs, *dir_cfgs;
  int i;

  svrcfg = ap_get_module_config(cmd->server->module_config, &snake_module);
  command_recs = svrcfg->command_recs;

  /* Find the internal directive which matches up with the requested
     directive */
  int_dir = mod_snake_internal_directive_find(command_recs, cmd->cmd->name);
  if(int_dir == NULL)
    return ap_psprintf(cmd->pool, "mod_snake: Found no module to process "
		       "dynamic directive: %s", cmd->cmd->name);

  if(mod_snake_get_dir_cback_data(per_dir_vdata, int_dir->pymod) == -1)
    return "mod_snake:  Error setting up per-dir config data!";


  /* Setup a new threadstate for us to have a working context for calling */
  /* Lock */
  *new_state = mod_snake_create_thread_and_lock(int_dir->pymod->interp);
  if(*new_state == NULL)
    return "mod_snake:  Failed to create a new interpreter state!";

  *callback_data = Py_None;
  dir_cfgs = dircfg->dir_config_data;
  for(i=0; i< dir_cfgs->nelts; i++){
    ModSnakeCallbackData *cbdata = (ModSnakeCallbackData *)dir_cfgs->elts + i;
    
    if(cbdata->module == int_dir->pymod){
      *callback_data = cbdata->data;
      break;
    }
  }

  *callback_func = int_dir->funcptr;
  *pymod = int_dir->pymod;
  return NULL;
}

static const 
char *mod_snake_internal_directive_no_args(cmd_parms *cmd, void *per_dir_vdata)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *cb_data, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						&cb_data, &func, &pymod);
  if(cres) return cres;

  args = Py_BuildValue("(O)", cb_data);  /* Incref */
  pymod->current_hook = SNAKE_HOOK_DIRECTIVE;
  res = PyEval_CallObject(func, args);
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
		      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;

  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
  return cres;
}

static const 
char *mod_snake_internal_directive_take1(cmd_parms *cmd, void *per_dir_vdata,
					 const char *arg1)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *cb_data, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						&cb_data, &func, &pymod);
  if(cres) return cres;

  if(cmd->cmd->args_how == FLAG)
    args = Py_BuildValue("(Oi)", cb_data, (int)arg1);
  else
    args = Py_BuildValue("(Os)", cb_data, arg1);  /* Incref */
  pymod->current_hook = SNAKE_HOOK_DIRECTIVE;
  res = PyEval_CallObject(func, args);
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
		      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;
  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
  return cres;
}

static const 
char *mod_snake_internal_directive_take2(cmd_parms *cmd, void *per_dir_vdata,
					 const char *arg1, const char *arg2)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *cb_data, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
						&cb_data, &func, &pymod);
  if(cres) return cres;

  args = Py_BuildValue("(Oss)", cb_data, arg1, arg2);  /* Incref */
  pymod->current_hook = SNAKE_HOOK_DIRECTIVE;
  res = PyEval_CallObject(func, args);
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
		      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;
  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */

  return cres;
}


static const 
char *mod_snake_internal_directive_take3(cmd_parms *cmd, void *per_dir_vdata,
					 const char *arg1, const char *arg2,
					 const char *arg3)
{
  ModSnakePyMod *pymod;
  PyThreadState *my_state;
  PyObject *func, *cb_data, *args, *res;
  const char *cres;

  /* Lock */
  cres = mod_snake_setup_internal_directive_call(cmd, per_dir_vdata, &my_state,
					        &cb_data, &func, &pymod);
  if(cres) return cres;

  args = Py_BuildValue("(Osss)", cb_data, arg1, arg2, arg3);  /* Incref */
  pymod->current_hook = SNAKE_HOOK_DIRECTIVE;
  res = PyEval_CallObject(func, args);
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(args);
  if(res == NULL){
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
    return ap_psprintf(cmd->pool, "mod_snake: Error calling command '%s'",
		      cmd->cmd->name);
  }

  cres = PyString_Check(res) ? ap_pstrdup(cmd->pool, PyString_AsString(res))
                             : NULL;
  Py_DECREF(res);
  mod_snake_destroy_thread_and_unlock(my_state);  /* Unlock */
  return cres;
}


/*
 * mod_snake_add_directive:  This routine is a haaaaaaaack.  It allows Python
 *                           modules to create their own directives that
 *                           can be used in Apache configuration.  Each 
 *                           module created adds its own callback information
 *                           to the directory config.  When the dir config
 *                           is destroyed, the reference to the modules
 *                           callback is also destroyed.
 *
 * Arguments:                pymod = Module associated with this directive.
 *                           directive_name = Directive string
 *                           location  = overrides information
 *                           functype  = TAKE1, etc.
 *                           funcptr   = Python func to call on directive
 *                           func_desc = Description of the function
 *
 * Return values:            Returns -1 on error, 0 on success.
 */

int mod_snake_add_directive(ModSnakePyMod *pymod,
			    const char *directive_name, int location, 
			    int functype, PyObject *funcptr, 
			    const char *func_desc)
{
  char *loc_directive_name = NULL, *loc_func_desc = NULL;
  ModSnakeIntDirective *intdir_add;
  ap_array_header_t *intdir_cfgs;
  ModSnakeServerCfg *svrcfg = pymod->parentsvr;
  command_rec *recs;
  int howmany;

  /* Check to see if this is a redefinition within the same config space */
  intdir_cfgs = svrcfg->command_recs;
  if(mod_snake_internal_directive_find(intdir_cfgs, directive_name) != NULL)
    return -1;

  /* First, count how many existing commands we have, in addition to checking
     for duplications. */
  recs = (command_rec *)snake_module.cmds;
  while(recs->name != NULL){
    if(!strcasecmp(recs->name, directive_name)){
      /* This is OK.  It simply means that we don't need to grow the 
	 command_rec structure for some new directive. Assume that the
	 configuration for this new entity (i.e. location, etc) is the
	 same as the new addition */

      Py_INCREF(funcptr);
      intdir_add          = ap_push_array(svrcfg->command_recs);
      intdir_add->name    = ap_pstrdup(svrcfg->cfgpool, directive_name);
      intdir_add->funcptr = funcptr;
      intdir_add->pymod   = pymod;
      return 0;
    }
    recs++;
  }


  howmany = recs - snake_module.cmds + 1; /*  include trailing NULL rec */

  /* Use regular malloc here, as we don't want anything to go away when
     a server config is destroyed. 
     XXX this will probably make us grow on graceful restarts, etc. */

  recs = NULL;
  
  if(func_desc == NULL)
    func_desc = "Unknown parameters";

  if((recs = malloc(sizeof(mod_snake_cmds[0]) * (howmany + 1))) == NULL ||
     (loc_directive_name = strdup(directive_name)) == NULL ||
     (loc_func_desc = strdup(func_desc)) == NULL)
  {
    /* free(NULL) ok, by ANSI */
    free(recs);
    free(loc_directive_name);
    free(loc_func_desc);
    return -1;
  }
  
  memcpy(recs, snake_module.cmds, sizeof(mod_snake_cmds[0]) * (howmany));
  if(snake_module.cmds != mod_snake_cmds)
     free((command_rec *)snake_module.cmds);  /* Free up old allocs */

  Py_INCREF(funcptr);
  intdir_add          = ap_push_array(svrcfg->command_recs);
  intdir_add->name    = ap_pstrdup(svrcfg->cfgpool, directive_name);
  intdir_add->funcptr = funcptr;
  intdir_add->pymod   = pymod;
  
  (recs + howmany - 1)->name         = loc_directive_name;
  (recs + howmany - 1)->cmd_data     = NULL;
  (recs + howmany - 1)->req_override = location;
  (recs + howmany - 1)->args_how     = functype;
  (recs + howmany - 1)->errmsg       = loc_func_desc;
  switch(functype){
  case NO_ARGS:
    (recs + howmany - 1)->func = mod_snake_internal_directive_no_args; break;
  case TAKE2:
  case ITERATE2:
  case TAKE12:
    (recs + howmany - 1)->func = mod_snake_internal_directive_take2; break;
  case TAKE3:
  case TAKE23:
  case TAKE123:
  case TAKE13:
    (recs + howmany - 1)->func = mod_snake_internal_directive_take3; break;
  case TAKE1:
  case ITERATE:
  case FLAG:
  case RAW_ARGS:
  default:
    (recs + howmany - 1)->func = mod_snake_internal_directive_take1;
  }

  (recs + howmany)->name             = NULL;

  snake_module.cmds = recs;
  return 0;
}

/* mod_snake_log_error: routine that we can call that will call the
 * 			appropriate ap_log_error function.
 *
 * Arguments:		Same as ap_log_error		
 *
 * Returns:		nothing
 */

void mod_snake_log_error(const char *file, int line, int level, ap_status_t status,
		const server_rec *s, const char *fmt, ...)
{
	va_list args;
        char big_ass_buff[2048];

	va_start(args, fmt);
	ap_vsnprintf(big_ass_buff, sizeof(big_ass_buff)-1, fmt, args);
#ifdef APACHE13
	ap_log_error(file, line, level, s, big_ass_buff);
#else /* Apache 2.0 */
	ap_log_error(file, line, level, status, s, big_ass_buff);
#endif
	 va_end(args);
	return;
}
