/* ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *     the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#ifndef MOD_SNAKE_DOT_H
#define MOD_SNAKE_DOT_H

#include "config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
 
#ifndef APACHE13
#include "http_connection.h"
#endif

#include "http_protocol.h"
#include "http_request.h"
#include "Python.h"
#include "snake_precompile.h"
#include <sys/param.h>

#ifdef APACHE13
#define APR_SUCCESS 0
#define ap_pool_t pool
#define ap_context_t pool
#define ap_array_header_t array_header
#define ap_table_t table
#define ap_table_entry_t table_entry
#define ap_stat stat
#define ap_size_t size_t
#define ap_ssize_t ssize_t
#define ap_time_t time_t
#define ap_status_t int
#endif

typedef struct mod_snake_server_cfg_st {
  ap_pool_t         *cfgpool;        /* The pool the cfg was created with   */
  ap_array_header_t *moduledirs;     /* Paths to load for all the modules   */
  ap_array_header_t *modules;        /* Modules loaded (ModSnakePyMod)      */
  ap_array_header_t *command_recs;   /* Array of ModSnakeIntDirective's.
					Used to keep track for each module 
					registering its own directives      */
  server_rec        *server;         /* Server which created us             */
} ModSnakeServerCfg;

typedef struct mod_snake_dir_cfg_st {
  ap_pool_t         *cfgpool;
  ap_array_header_t *dir_config_data; /* Array of ModSnakeCallbackData with
					 all the callback datas for the 
					 modules within this dir            */
  char              *path;
  struct mod_snake_dir_cfg_st  *parent; /* If directory configurations are 
					 merged, this points to the base
					 dir-cfg that is generated from the
					 (base, new) pair */
} ModSnakeDirCfg;


#define SNAKE_HOOK_DIRECTIVE                -3  /* Directive callbacks       */
#define SNAKE_HOOK_INIT                     -2  /* Initialization - fake hook*/
#define SNAKE_HOOK_NO_HOOK                  -1  /* Not a hook.. ;-)          */
#define SNAKE_HOOK_CREATE_DIR_CONFIG         0
#define SNAKE_HOOK_MERGE_DIR_CONFIG          1
#define SNAKE_HOOK_OPEN_LOGS                 2
#define SNAKE_HOOK_HEADER_PARSER             3
#define SNAKE_HOOK_POST_READ_REQUEST         4
#define SNAKE_HOOK_LOG_TRANSACTION           5
#define SNAKE_HOOK_TRANSLATE_NAME            6
#define SNAKE_HOOK_CHECK_USER_ID             7
#define SNAKE_HOOK_FIXUPS                    8
#define SNAKE_HOOK_TYPE_CHECKER              9
#define SNAKE_HOOK_ACCESS_CHECKER            10
#define SNAKE_HOOK_AUTH_CHECKER              11
#define SNAKE_HOOK_CONTENT_HANDLER           12
#define SNAKE_HOOK_PRE_CONNECTION            13
#define SNAKE_HOOK_PROCESS_CONNECTION        14
#ifdef APACHE13
#define SNAKE_HOOK_NUM_HOOKS                 13  /* Update when above changes*/
#else
#define SNAKE_HOOK_NUM_HOOKS                 15  /* Update when above changes*/
#endif

#define SNAKE_HOOK_BOFFO       -1     /* Means that the hook really doesn't
					 have an order type */
#define SNAKE_HOOK_FIRST        0     
#define SNAKE_HOOK_ALL          1
#define SNAKE_HOOK_VOID         2

typedef struct mod_snake_pymod_st {
  ModSnakeServerCfg *parentsvr;    /* Parent server cfg (if server owns it) */
  char *name;                      /* As per the config directive           */

  /* The below PyObjects are only referenced at most 1 time from the
     main program.  */
  PyObject *module;                /* Loaded python module (from import)    */
  PyObject *mod_obj;               /* Instantiation of the object           */
  PyInterpreterState *interp;      /* Interpreter the threads run in for
				      this module                           */
  PyThreadState  *main_state;      /* State in which the interp was created */
  int failexit;                    /* 1 if a failure within this module
				      causes a server exit                  */
  int valid;                       /* Whether or not the module is actually
				      loaded.  If the module IS loaded
				      then the above PyObjects will be 
				      non-NULL                              */
  /* Hooks */
  int current_hook;                /* Designates which hook is currently
				      being called out to.  -1 indicates
				      no hook being called                  */
  PyObject *hooks[SNAKE_HOOK_NUM_HOOKS];
} ModSnakePyMod;

typedef struct mod_snake_callback_data_st {
  PyObject *data;          /* Data the module expects as a callback */
  ModSnakePyMod *module;   /* Module associated with this data      */
} ModSnakeCallbackData;


typedef struct mod_snake_internal_directive_st {
  PyObject *funcptr;       /* Actual object to call                  */
  ModSnakePyMod *pymod;    /* Module owning this directive           */
  char *name;              /* Name of the directive this matches     */
} ModSnakeIntDirective;

/* Swig-made modules */
extern void initmod_snakec();
extern void initaprc();
extern PyObject *mod_snake_get_generic_shadowclass(void *obj, 
						   PyObject *(*func)(void *),
						   char *sclass);
extern PyObject *mod_snake_get_ap_table_t_shadowclass(ap_table_t *table);
extern PyObject *mod_snake_get_conn_rec_shadowclass(conn_rec *c);
extern PyObject *mod_snake_get_server_rec_shadowclass(server_rec *s);
extern PyObject *mod_snake_get_server_addr_rec_shadowclass(server_addr_rec *s);
extern PyObject *mod_snake_get_request_rec_shadowclass(request_rec *rec);
extern PyObject *mod_snake_get_modsnakepymod_shadowclass(ModSnakePyMod *pymod);
extern PyObject *mod_snake_get_require_line_shadowclass(require_line *reqline);

extern int mod_snake_add_directive(ModSnakePyMod *pymod,
				   const char *directive_name, int location, 
				   int functype, PyObject *funcptr, 
				   const char *func_desc);

extern int mod_snake_initmod_internal(char *modname, SnakePrecompiledPy *);
extern int mod_snake_setup_syspath(ap_array_header_t *moddirs);
extern void mod_snake_unload_module(ModSnakePyMod *pymod);
extern int mod_snake_import_module(ModSnakePyMod *pymod, server_rec *s);
extern int mod_snake_load_module(ModSnakePyMod *pymod, 
				 ap_array_header_t *syspath, server_rec *s);

extern PyThreadState *mod_snake_create_thread_and_lock(PyInterpreterState *);
extern void mod_snake_destroy_thread_and_unlock(PyThreadState *tstate);

extern PyObject * 
mod_snake_run_hook_merge_dir_config(ModSnakePyMod *pymod, server_rec *s,
				    PyObject *base_cdata, PyObject *new_cdata,
				    PyObject *callback);

extern PyObject * 
mod_snake_run_hook_create_dir_config(ModSnakePyMod *pymod, server_rec *s,
				     ModSnakeDirCfg *dir_cfg,
				     PyObject *callback);

extern 
int mod_snake_run_generic_request_rec(ModSnakePyMod *pymod, request_rec *r,
				      ModSnakeDirCfg *per_dir_cfg,
				      PyObject *callback, int cb_num);
extern 
int mod_snake_run_generic_conn_rec(ModSnakePyMod *pymod, conn_rec *c,
				   PyObject *callback, int cb_num);

extern int mod_snake_hooknum_to_type(int hookno);
extern int mod_snake_hookname_to_num(const char *hookname);
extern const char *mod_snake_hooknum_to_name(int hookno);
extern void mod_snake_run_hook_open_logs(ModSnakePyMod *pymod);

extern void mod_snake_log_error(const char *file, int line, int level,
		ap_status_t status, const server_rec *s, const char *fmt, ...)
#ifdef __GNUC__
		__attribute__((format(printf,6,7)))
#endif
     ;
#endif
