/* ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *     the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <Python.h>
#include <marshal.h>

static
int process_file(const char *infile, const char *outfile, const char *cname){
  FILE *infp = NULL, *outfp = NULL;
  char py_cname[1024];
  char *inbuf, *cp;
  struct stat instat;
  PyObject *code, *codestr;
  int i;

  if(stat(infile, &instat) == -1){
    fprintf(stderr, "Error stat()ing '%s': %s\n",
	    infile, strerror(errno));
    return -1;
  }

  if((inbuf = malloc(instat.st_size + 1)) == NULL){
    fprintf(stderr, "Error allocating: %ld bytes\n", (long)instat.st_size);
    return -1;
  }

  if((infp = fopen(infile, "r")) == NULL ||
     (outfp = fopen(outfile, "wb")) == NULL)
  {
    fprintf(stderr, "Error opening '%s' for %s: %s\n", 
	    (infp == NULL) ? infile : outfile,
	    (infp == NULL) ? "read" : "write",
	    strerror(errno));
    return -1;
  }

  if(fread(inbuf, 1, instat.st_size, infp) != instat.st_size){
    fprintf(stderr, "Error reading from '%s': %s\n",
	    infile, strerror(errno));
    fclose(infp);
    fclose(outfp);
    free(inbuf);
    return -1;
  }
  fclose(infp);

  inbuf[instat.st_size] = '\0';
  sprintf(py_cname, "<internal %s>", cname);
  if((code = Py_CompileString(inbuf, py_cname, Py_file_input)) == NULL ||
     (codestr = PyMarshal_WriteObjectToString(code)) == NULL)
  {
    fprintf(stderr, "Error compiling python code\n");
    PyErr_Print();
    Py_XDECREF(code);
    Py_XDECREF(codestr);
    fclose(outfp);
    free(inbuf);
    return -1;
  }
  free(inbuf);
  Py_DECREF(code);

  fprintf(outfp, "#include \"snake_precompile.h\"\n\n");
  fprintf(outfp, "static char local%s[] = {\n", cname);
  cp = PyString_AsString(codestr);
  for(i = 0; i < PyString_Size(codestr); i++, cp++){
    fprintf(outfp, "%4d,", (int)*cp);
    if(!((i+1) % 15)) fprintf(outfp, "\n");
  }

  fprintf(outfp, "};\n");
  fprintf(outfp, "const SnakePrecompiledPy %s = {\n", cname);
  fprintf(outfp, "local%s, sizeof(local%s)\n };\n", cname, cname);
  Py_DECREF(codestr);
  fclose(outfp);
  return 0;
}

int main(int argc, char *argv[]){
  Py_Initialize();

  if(argc != 4){
    fprintf(stderr, "Syntax: %s <infile> <outfile> <out_cname>\n", argv[0]);
    return -1;
  }

  return process_file(argv[1], argv[2], argv[3]);
}
